import x10.util.Timer;
import x10.io.Console;

class laplacian {

    public static def println(s:String) = Console.ERR.println(s);
    public static def print(s:String) = Console.ERR.print(s);
    public static def now():double = Timer.nanoTime() * 1e-9;
    public static def format(x:double, numDecimals:int) {
	return String.format("%1."+numDecimals+"f", [x as Box[Double]]);
    }

    global u : Array[double](2);
    global v : Array[double](2);
    global n : Int, l : Int;
    global lambda : Double;

    public def this(m:Int)
    {
        var k : int = m/Place.MAX_PLACES;
	var R : Region(2) = [0..(k+2)*Place.MAX_PLACES-1,0..m+1];
	var D : Dist(2) = Dist.makeBlock(R, 0);
	u = Array.make[double](D, (p:Point) => 
	    { ((here.id() == 0) && (p(0) == 0)) ? 1.0 : 0.0 });
	v = Array.make[double](D, (p:Point) => 
	    { ((here.id() == 0) && (p(0) == 0)) ? 1.0 : 0.0 });
        n = m;
        l = k;
        lambda = +0.025;
    }

    public def output()
    {
	println("");
	for (var i:int=0; i<Place.MAX_PLACES; i++) {
	    val ii=i;
	    at (Place.place(i)) {
		j0:int = ii*(l+2);
		j1:int = j0+l+2;
		for (var j:int=j0; j<j1; j++) {
		    for (var k:int=0; k<n+2; k++)
			print(" " + format(u(j, k), 6));
		    println("");
		}
	    }
	    println("");
	}
    }

    public def iterate()
    {
       finish
	for (var i:int=0; i<Place.MAX_PLACES; i++)  {
           val ii = i;
           async (Place.place(i)) {
	       val j0:int = ii*(l+2)+1;
	       val j1:int = j0+l;
	       foreach ((j,k) : Point in [j0..j1-1, 1..n]) {
		       v(j,k) = (1-4*lambda)*u(j,k) + lambda* 
			   ( u(j+1, k) + 
			     u(j-1, k) + 
			     u(j, k-1) +
			     u(j, k+1)); 
		   }
	   }
	}

        val d = Array.make[double](Dist.makeUnique(), (Point)=>0.0);
	finish
	for (var i:int=0; i<Place.MAX_PLACES; i++)  {
	    val ii = i;
	    async (Place.place(i)) {
		var dd : Double = 0.0;
		j0:int = ii*(l+2)+1;
		j1:int = j0+l;
		for (var j:int=j0; j<j1; j++) {
		    for (var k:int=1; k<n+1; k++) {
			dd = Math.abs(u(j,k) - v(j,k));
			if (dd > d(ii)) d(ii) = dd;
			u(j,k) = v(j,k);
		    }
		}
	    }
        }
	
	return d.reduce(Math.max.(double,double), 0.0);
  	
    }

    public def exchange()
    {
	for (var i:int=1; i<Place.MAX_PLACES; i++) {
	    val ii = i;
	    val i0 = ii*(l+2);
	    val v = at (Place.places(ii-1)) 
		ValRail.make[double](n, (j:Int)=>u(i0-2,j+1));
	    val w = at (Place.places(ii)) 
		ValRail.make[double](n, (j:Int)=>u(i0+1,j+1));

	    at (Place.places(ii-1)) 
		for (var j:int=0; j<n; j++)
		    u(i0-1,j+1) = w(j);
	    
	    at (Place.places(ii)) 
		for (var j:int=0; j<n; j++)
		    u(i0,j+1) = v(j);
	}
    }

    public static def main(args:Rail[String]!) {

	var n:Int = (args.length > 0) ? int.parseInt(args(0)) : 10;
	var it:Int = 0;
        
        var err:Double = 1.0;

        val l = new laplacian(n);

        var i:Int;
        if (n <= 10)
           l.output();
	
	var t:double = now();

	while ((err > 1e-3) && (it < 10)) {
	    it = it + 1;
            err = l.iterate();
	    
	    l.exchange();
	    println("it = " + it + " err = " + err);
	}
	println("time elapsed : " + (now() - t));

        if (n <= 10)
	    l.output();
	
   }

}
