import x10.util.Timer;
import x10.io.Console;

class laplacian {

    public static def println(s:String) = Console.ERR.println(s);
    public static def print(s:String) = Console.ERR.print(s);
    public static def now():double = Timer.nanoTime() * 1e-9;
    public static def format(x:double, numDecimals:int) {
	return String.format("%1."+numDecimals+"f", [x as Box[Double]]);
    }

    global u : Array[double](2);
    global v : Array[double](2);
    val n : Int, l : Int;
    global lambda : Double;

    public def this(m:Int, k:Int)
    {
	n = m;
        l = k;
        lambda = +0.025;

	var R : Region(2) = [0..k+1,0..m+1];
	var D : Dist(2) = R->here;
	u = Array.make[double](D, (p:Point) => 
	    { ((here.id() == 0) && (p(0) == 0)) ? 1.0 : 0.0 });
	v = Array.make[double](D, (p:Point) => 
	    { ((here.id() == 0) && (p(0) == 0)) ? 1.0 : 0.0 });
    }

    public def output()
    {
	var i:int;
	var j:int;

	println("");
	for (i=0; i<l+2; i++) {
	    for (j=0; j<n+2; j++) {
	       print(" " + format(u(i, j), 6));
	    }
	    println("");
	}
    }

    public def iterate()
    {
	var i:int;
	var j:int;
	for (i=1; i<l+1; i++) {
	    for (j=1; j<n+1; j++) {
	       v(i,j) = (1-4*lambda)*u(i,j) + lambda* 
		                   ( u(i+1, j) + 
		                     u(i-1, j) + 
		                     u(i, j-1) +
		                     u(i, j+1)); 
	    }
	}

	var dmax : Double = -1.0;
	var d : Double = 0.0;
	for (i=1; i<l+1; i++) {
	    for (j=1; j<n+1; j++) {
		d = Math.abs(u(i,j) - v(i,j));
		if (d > dmax) dmax = d;
		u(i,j) = v(i,j);
	    }
	} 
	return dmax;
  	
    }

    public static def exchange(val l: Array[laplacian](1)!)
    {

	for (var i:int=0; i<Place.MAX_PLACES-1; i++) {
	    val i0 = i;
	    val i1 = i+1;
	    val v = at (Place.places(i0)) 
		ValRail.make[double](l(i0).n+2, 
				     (j:Int)=>l(i0).u(l(i0).l,j));
	    val w = at (Place.places(i1)) 
		ValRail.make[double](l(i1).n+2, 
				     (j:Int)=>l(i1).u(1,j));

	    at (Place.places(i0)) 
		for (var i:int=0; i<=l(i0).n; i++)
		    l(i0).u(l(i0).l+1,i) = w(i);
	    
	    at (Place.places(i1)) 
		for (var i:int=0; i<=l(i1).n; i++)
		    l(i1).u(0,i) = v(i);
	}
    }

    public static def main(args:Rail[String]!) {

	val n:Int = (args.length > 0) ? int.parseInt(args(0)) : 10;
        val m:Int = n/Place.MAX_PLACES;

	var it:Int = 0;
	var d : Dist(1) = Dist.makeUnique();

	val err = Array.make[Double](d);
        var err_glob:Double = 1.0;

        val l = Array.make[laplacian](d, (p:Point) => new laplacian(n, m));

        var i:Int;
        if (n <= 10)
	for (i=0; i<Place.MAX_PLACES; i++)
            at (Place.places(i)) l(here.id()).output();

	var t:double = now();

	while ((err_glob > 1e-3) && (it < 10)) {
	    it = it + 1;
            finish
	        for (i=0; i<Place.MAX_PLACES; i++)
                    async (Place.places(i)) 
			{ err(here.id()) = l(here.id()).iterate();}

	    exchange(l);

	    err_glob = 0.0;
	    for (p:Place in Place.places) {
                val e = at (p) err(here.id());
                if (e > err_glob) err_glob = e;
            }
	    println("it = " + it + " err = " + err_glob);
	}
	println("time elapsed : " + (now() - t));

        if (n <= 10)
	for (i=0; i<Place.MAX_PLACES; i++)
            at (Place.places(i)) l(here.id()).output();
   }

}
